//SINGLE ROUND MODULE FOR ENCRYPTION


module encryption_single(data_set,clock,reset,expanded_key,round_num,round_data,round_over);

input reg   [7:0] data_set [3:0][3:0];     //INPUT DATA SET REPRESENTED IN MATRIX FORM
input clock;
input reg   reset;
input reg   [7:0] expanded_key [3:0][3:0]; //KEY GENERATED FORM KEY GENERATOR MODULE
input reg [3:0] round_num;                 //ROUND NUMBER
output reg [7:0] round_data [3:0][3:0];    //OUTPUT DATA FROM THE ROUND
output reg round_over;                     //STATUS SIGNAL TO INIDICATE ROUND IS OVER


reg [7:0] data_set_latch [3:0][3:0];
reg [7:0] state_matrix0 [3:0][3:0];
reg [7:0] state_matrix1 [3:0][3:0];
reg [7:0] state_matrix2 [3:0][3:0];
reg [7:0] state_matrix3 [3:0][3:0];
reg [7:0] state_matrix3a [3:0][3:0][3:0];
reg [7:0] state_matrix3b [3:0][3:0][3:0];
reg [7:0] state_matrix3c [3:0][3:0][3:0];
reg [7:0] state_matrix4 [3:0][3:0];

reg [7:0] data_out    [3:0][3:0];

reg [7:0] mix_matrix    [3:0][3:0];

reg [7:0] addr [3:0];
reg [7:0] data [3:0];

reg r_done;

reg enable;
reg [5:0] state;
reg [4:0] round,look_count;

reg [4:0] i,j,k;



parameter init1 = 5'd0, init2 = 5'd1, init3 = 5'd2, init4 = 5'd3, init5 = 5'd4, round2b = 5'd5, round2a = 5'd6, round2 = 5'd7, round3a = 5'd8, round3d = 5'd9, round4 = 5'd10, done1 = 5'd11, done2 = 5'd12, dummy1 = 5'd13, dummy2 = 5'd14;


//4 LOOK UP TABLES FOR ACCESSIG THE S-BOX AND PERFORMING THE SUBSTITUTION 4 CHARACTERS AT A TIME
look_up L0(
			.clk(clock),
			.enable(enable),
			.in(addr[0]),
			.out(data[0])
			 );

look_up L1(
			.clk(clock),
			.enable(enable),
			.in(addr[1]),
			.out(data[1])
			 );

look_up L2(
			.clk(clock),
			.enable(enable),
			.in(addr[2]),
			.out(data[2])
			 );
			 
look_up L3(
			.clk(clock),
			.enable(enable),
			.in(addr[3]),
			.out(data[3])
			 );

			 

reg [5:0] n1,n2,n3;
always@(posedge clock) begin
if(reset) begin

	for(i=0;i<4;i=i+1) begin
		for(j=0;j<4;j=j+1) begin
			state_matrix1[i][j] <= 0;
			state_matrix2[i][j] <= 0;
			state_matrix3[i][j] <= 0;
			state_matrix4[i][j] <= 0;
			data_set_latch[i][j] <= data_set[i][j];
			for(k=0;k<4;k=k+1) begin
				state_matrix3a[i][j][k] <= 0;
				state_matrix3b[i][j][k] <= 0;
				state_matrix3c[i][j][k] <= 0;
	      end
		end
	end
	
	enable<=0;
	state<= init1;
	round<=0;
	r_done <= 0;

//MIX MATRIX	
	mix_matrix[0][0] <= 8'd2;
	mix_matrix[0][1] <= 8'd3;
	mix_matrix[0][2] <= 8'd1;
	mix_matrix[0][3] <= 8'd1;
	
	mix_matrix[1][0] <= 8'd1;
	mix_matrix[1][1] <= 8'd2;
	mix_matrix[1][2] <= 8'd3;
	mix_matrix[1][3] <= 8'd1;
	
	mix_matrix[2][0] <= 8'd1;
	mix_matrix[2][1] <= 8'd1;
	mix_matrix[2][2] <= 8'd2;
	mix_matrix[2][3] <= 8'd3;
	
	mix_matrix[3][0] <= 8'd3;
	mix_matrix[3][1] <= 8'd1;
	mix_matrix[3][2] <= 8'd1;
	mix_matrix[3][3] <= 8'd2;
	n2 <= 0;
	n1 <= 0;
	
end

else begin
	case(state) 
//INITIAL STALLING TO PREVENT TIMING INCONSISTENIES WHEN THE ROUNDS ARE PIPELINED
	init1: begin

			state <= init2;
			n2 <= n1+1;
			
		end
		
		init2 : begin
			state <= init3;
			n1 <= 4'd3;
		end
		
		init3 : begin
			state <= init4;
			n1 <= 4'd5;
		end
		
		init4 : begin
			state <= init5;
		end
		
		init5 : begin
			state <= round2b;
			enable<=1;
			look_count=0;
		end


//COMBINATION OF ROUND 1 AND 2 TOGETHER		
		round2b : begin
			for(k=0;k<4;k=k+1) begin
				addr[k] <= data_set_latch[0][k];
			end
			state <= round2a;
		end

		round2a: begin
			state <= round2;
		end
		
		round2: begin
			if(look_count==0) begin
				for(i=0;i<4;i=i+1) begin
					state_matrix2[0][i] <= data[i];
				end
				addr[0] <= data_set_latch[1][1];
				addr[1] <= data_set_latch[1][2];
				addr[2] <= data_set_latch[1][3];
				addr[3] <= data_set_latch[1][0];
				state <= round2a;
			end
			
			else if(look_count == 1) begin
				for(i=0;i<4;i=i+1) begin
					state_matrix2[1][i] <= data[i];
				end
				addr[0] <= data_set_latch[2][2];
				addr[1] <= data_set_latch[2][3];
				addr[2] <= data_set_latch[2][0];
				addr[3] <= data_set_latch[2][1];
				state <= round2a;
			end
			
			else if(look_count == 2) begin
				for(i=0;i<4;i=i+1) begin
					state_matrix2[2][i] <= data[i];
				end
				addr[0] <= data_set_latch[3][3];
				addr[1] <= data_set_latch[3][0];
				addr[2] <= data_set_latch[3][1];
				addr[3] <= data_set_latch[3][2];
				state <= round2a;
			end
			
			else begin
				for(i=0;i<4;i=i+1) begin
					state_matrix2[3][i] <= data[i];
				end

//IF THE ROUND NUMBER IS 10, THEN AVOID THE THIRD STEP AND PROCEED TOWARDS THE DUMMY STAGE AND THEN TO THE FOURTH STEP
				if(round_num == 4'd10) begin
					state <= dummy1;
				end
				else begin
					state <= round3a;
				end
			end
			
			look_count = look_count + 1;
		end
	
//DUMMY STAGES TO ENSURE THAT ALL ROUNDS HAVE SAME EXECUTION TIME
		dummy1 : begin
			state <= dummy2;
		end
		
		dummy2 : begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					state_matrix3[i][j] <= state_matrix2[i][j];
				end
			end
			state <= round4;
		end
		
//STEP 3: MULTIPLICATION WITH MIX MATRIX		
		round3a: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					for(k=0;k<4;k=k+1) begin
						if(mix_matrix[i][k]==1) begin
							state_matrix3c[i][j][k] <= state_matrix2[k][j];
						end
						else if(mix_matrix[i][k] == 2) begin
							if(state_matrix2[k][j] < 8'd128) begin
								state_matrix3c[i][j][k] <= state_matrix2[k][j]<<1;
							end
							else begin
								state_matrix3c[i][j][k] <= ((state_matrix2[k][j])<<1)^(8'd27);
							end
						end
						else if(mix_matrix[i][k] == 3) begin
							if(state_matrix2[k][j] <8'd128) begin
								state_matrix3c[i][j][k] <= ((state_matrix2[k][j])<<1)^(state_matrix2[k][j]);
							end
							else begin
								state_matrix3c[i][j][k] <= ((state_matrix2[k][j])<<1)^(8'd27)^(state_matrix2[k][j]);
							end
						end
					end
				end
			end
			state <= round3d;
		end
		
		round3d: begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					state_matrix3[i][j] <= state_matrix3c[i][j][0]^state_matrix3c[i][j][1]^state_matrix3c[i][j][2]^state_matrix3c[i][j][3];
				end
			end
			state <= round4;
		end
	

//STEP4 : XOR OPERATION WITH THE KEY GENERATED	
		round4 : begin
			for(i=0;i<4;i=i+1) begin
				for(j=0;j<4;j=j+1) begin
					state_matrix4[i][j] <= state_matrix3[i][j]^expanded_key[i][j];
				end
			end
			
			state <= done1;
		end
		
		done1 : begin
			r_done <= 1;
			state <= done2;
		end
		
		done2 : begin
			r_done <= 0;
		end
		

	endcase
end
end

		
		
assign round_data = state_matrix4;
assign round_over = r_done;

endmodule

